from typing import cast, Optional, Union
from hexbytes import HexBytes

from web3 import Web3, AsyncWeb3
from web3.contract.contract import (
    ContractFunction,
    ContractFunctions,
    ContractCaller,
)
from web3.contract.base_contract import (
    NonExistentFallbackFunction,
    NonExistentReceiveFunction,
)
from web3._utils.datatypes import PropertyCheckingFactory
from web3._utils.abi import (
    filter_by_type,
    fallback_func_abi_exists,
    receive_func_abi_exists,
)
from web3._utils.function_identifiers import FallbackFn, ReceiveFn
from web3.exceptions import ContractLogicError
from web3.types import ABI, TxReceipt
from eth_typing import ChecksumAddress
import eth_account

from cheb3.account import Account
from cheb3.constants import GAS_BUFFER

from loguru import logger


class Contract:
    """Please use :func:`cheb3.Connection.contract` interface to
    create a contract instance associated with the connection.
    """

    # set during class construction
    w3: Web3 = None

    def __init__(self, signer: Account = None, address: str = None, **kwargs) -> None:
        """
        You can read states from the blockchain even if `signer` is not given.
        """

        if self.w3 is None:
            raise AttributeError(
                "The `Contract` class has not been initialized. Please use the "
                "`Connection.contract` interface to create a contract."
            )

        self.signer = signer.eth_acct if signer else None
        if address:
            self.address = address
            self.instance = self.w3.eth.contract(address=address, **kwargs)
            self._init_functions()
        else:
            self.address = None
            self.instance = self.w3.eth.contract(**kwargs)

    def deploy(self, *constructor_args, **kwargs) -> None:
        """Deploys the contract.

        :param constructor_args: Constructor arguments.

        Keyword Args:
            value (int): The amount to transfer, defaults to 0 (wei).
            gas_price (int): Specifies the gas price for the deployment.
            gas_limit (int): Specifies the maximum gas the deployment can use.
            proxy (bool): A minimal proxy contract (ERC-1167) will be deployed
                and connected to the logic contract if set to :const:`True`,
                defaults to :const:`False`.
            access_list (List[Dict]): Specify a list of addresses and storage
                keys that the transaction plans to access (EIP-2930). It will only
                be used in logic contract deployment if `proxy` is :const:`True`.
        """
        if not self.signer:
            raise AttributeError("The `signer` is missing.")

        if self.address:
            logger.info(f"Contract {type(self).__name__} has already been deployed at {self.address}.")
            return

        tx = self.signer.sign_transaction(
            self.instance.constructor(*constructor_args).build_transaction(
                {
                    "chainId": self.w3.eth.chain_id,
                    "nonce": self.w3.eth.get_transaction_count(self.signer.address),
                    "gas": kwargs.get(
                        "gas_limit",
                        self.instance.constructor(*constructor_args).estimate_gas({"from": self.signer.address}),
                    ) + GAS_BUFFER,
                    "gasPrice": kwargs.get("gas_price", self.w3.eth.gas_price),
                    "value": kwargs.get("value", 0),
                    "accessList": kwargs.get("access_list", []),
                }
            )
        ).rawTransaction
        logger.debug(f"Deploying {type(self).__name__} ...")
        tx_hash = self.w3.eth.send_raw_transaction(tx).hex()
        receipt = self.w3.eth.wait_for_transaction_receipt(tx_hash)
        if not receipt.status:
            raise Exception(f"Failed to deploy {type(self).__name__}.")
        logger.info(
            f"""The {"logic " if kwargs.get('proxy', False) else ""}{type(self).__name__} is deployed at {receipt.contractAddress}"""
        )
        self.address = receipt.contractAddress

        if kwargs.get("proxy", False):
            proxy_bytecode = f"3d602d80600a3d3981f3363d3d373d3d3d363d73{self.address[2:].lower()}5af43d82803e903d91602b57fd5bf3"
            tx = {
                "from": self.signer.address,
                "to": None,
                "chainId": self.w3.eth.chain_id,
                "nonce": self.w3.eth.get_transaction_count(self.signer.address),
                "gasPrice": kwargs.get("gas_price", self.w3.eth.gas_price),
                "data": proxy_bytecode,
            }
            tx["gas"] = kwargs.get("gas_limit", self.w3.eth.estimate_gas(tx) + GAS_BUFFER)
            tx = self.signer.sign_transaction(tx).rawTransaction
            logger.debug("Deploying the proxy ...")
            tx_hash = self.w3.eth.send_raw_transaction(tx).hex()
            receipt = self.w3.eth.wait_for_transaction_receipt(tx_hash)
            if not receipt.status:
                raise Exception("Failed to deploy the proxy.")
            logger.info(f"The proxy is deployed at {receipt.contractAddress}")
            self.address = receipt.contractAddress

        self.instance = self.w3.eth.contract(self.address, abi=self.instance.abi)
        self._init_functions()

    @staticmethod
    def get_fallback_function(abi: ABI, w3: Web3, signer: eth_account.Account, address: str):
        if abi and fallback_func_abi_exists(abi):
            return ContractFunctionWrapper.factory(
                "fallback",
                w3=w3,
                signer=signer,
                contract_abi=abi,
                address=address,
                function_identifier=FallbackFn,
            )()
        else:
            return cast(ContractFunctionWrapper, NonExistentFallbackFunction())

    @staticmethod
    def get_receive_function(abi: ABI, w3: Web3, signer: eth_account.Account, address: str):
        if abi and receive_func_abi_exists(abi):
            return ContractFunctionWrapper.factory(
                "receive",
                w3=w3,
                signer=signer,
                contract_abi=abi,
                address=address,
                function_identifier=ReceiveFn,
            )()
        else:
            return cast(ContractFunctionWrapper, NonExistentReceiveFunction())

    def _init_functions(self) -> None:
        self.functions = ContractFunctionsWrapper(self.signer, self.instance.abi, self.w3, self.address)
        self.caller = ContractCaller(self.instance.abi, self.w3, self.address)

        self.fallback = self.get_fallback_function(self.instance.abi, self.w3, self.signer, self.address)
        self.receive = self.get_receive_function(self.instance.abi, self.w3, self.signer, self.address)

    @classmethod
    def factory(cls, w3: Web3, contract_name: str = "") -> "Contract":
        contract = cast(
            Contract,
            PropertyCheckingFactory(contract_name or cls.__name__.lower(), (cls,), {"w3": w3}),
        )
        return contract

    def get_balance(self) -> int:
        """Returns the balance of the contract instance."""
        return self.w3.eth.get_balance(self.address)

    def get_storage_at(self, slot: int) -> HexBytes:
        """Returns the value from a storage position for the contract instance.

        :param slot: The storage slot.
        :type slot: int

        :rtype: ~hexbytes.main.HexBytes
        """
        return self.w3.eth.get_storage_at(self.address, slot)


class ContractFunctionsWrapper(ContractFunctions):
    def __init__(
        self,
        signer: eth_account.Account,
        abi: ABI,
        w3: Union["Web3", "AsyncWeb3"],
        address: Optional[ChecksumAddress] = None,
    ) -> None:
        self.abi = abi
        self.w3 = w3
        self.address = address
        self.signer = signer

        if self.abi:
            self._functions = filter_by_type("function", self.abi)
            for func in self._functions:
                setattr(
                    self,
                    func["name"],
                    ContractFunctionWrapper.factory(
                        func["name"],
                        w3=self.w3,
                        signer=self.signer,
                        contract_abi=self.abi,
                        address=self.address,
                        function_identifier=func["name"],
                    ),
                )


class ContractFunctionWrapper(ContractFunction):
    signer: eth_account.Account = None

    def send_transaction(self, **kwargs) -> TxReceipt:
        if not self.signer:
            raise AttributeError("The `signer` is missing.")

        try:
            estimate_gas = self.estimate_gas() + GAS_BUFFER
        except Exception:
            estimate_gas = 3000000
        tx = {
            "chainId": self.w3.eth.chain_id,
            "nonce": self.w3.eth.get_transaction_count(self.signer.address),
            "gasPrice": kwargs.get("gas_price", self.w3.eth.gas_price),
            "value": kwargs.get("value", 0),
            "gas": kwargs.get("gas_limit", estimate_gas),
        }
        if kwargs.get("access_list"):
            tx["accessList"] = kwargs["access_list"]
        tx = self.signer.sign_transaction(self.build_transaction(tx)).rawTransaction
        tx_hash = self.w3.eth.send_raw_transaction(tx).hex()
        func_name = self.function_identifier if isinstance(self.function_identifier, str) else self.function_identifier.__name__
        logger.info(f"({self.address}).{func_name} transaction hash: {tx_hash}")
        if not kwargs.get("wait_for_receipt", True):
            return tx_hash
        receipt = self.w3.eth.wait_for_transaction_receipt(tx_hash)
        if not receipt.status:
            raise Exception(f"Transact to ({self.address}).{func_name} errored.")
        return receipt



from typing import cast, Union
from hexbytes import HexBytes
import random
import string

from web3 import Web3
from web3._utils.datatypes import PropertyCheckingFactory
from web3.types import TxReceipt
from eth_typing import HexStr
import eth_account

from cheb3.constants import GAS_BUFFER

from loguru import logger


class Account:
    """Please use :func:`cheb3.Connection.account` interface to
    create an account instance associated with the connection.
    """

    w3: Web3 = None

    def __init__(self, private_key: str = None) -> None:
        """Creates a new account if `private_key` is not given."""

        if self.w3 is None:
            raise AttributeError(
                "The `Account` class has not been initialized. Please use the "
                "`Connection.account` interface to create an account."
            )

        if private_key is None:
            self.eth_acct = eth_account.Account.create("".join(random.choices(string.ascii_letters, k=3)))
        else:
            self.eth_acct = eth_account.Account.from_key(private_key)

        self.address = self.eth_acct.address
        self.private_key = self.eth_acct.key.hex()

    @classmethod
    def factory(cls, w3: Web3) -> "Account":
        eth_acct = cast(Account, PropertyCheckingFactory(cls.__name__, (cls,), {"w3": w3}))
        return eth_acct

    def get_balance(self) -> int:
        """Returns the balance of the account instance."""
        return self.w3.eth.get_balance(self.eth_acct.address)

    def call(self, to: HexStr, data: HexStr = "0x", **kwargs) -> HexBytes:
        """Interact with a smart contract without creating a new
        transaction on the blockchain.

        :param to: The address of the contract.
        :type to: HexStr
        :param data: The transaction data, defaults to `0x`.
        :type data: HexStr

        Keyword Args:
            state_override (dict): Specify the state override set.
                View `Geth documentation <https://geth.ethereum.org/docs/interacting-with-geth/rpc/ns-eth#eth-call>`_ for more details.

        :rtype: ~hexbytes.main.HexBytes
        """

        to = Web3.to_checksum_address(to)

        return self.w3.eth.call(
            {
                "to": to,
                "from": self.address,
                "data": data,
            },
            state_override=kwargs.get("state_override", None),
        )

    def send_transaction(
        self, to: Union[HexStr, None], value: int = 0, data: HexStr = "0x", **kwargs
    ) -> Union[TxReceipt, HexStr]:
        """Transfer ETH or interact with a smart contract.

        :param to: The address of the receiver.
        :type to: Union[HexStr, None]
        :param value: The amount to transfer, defaults to 0 (wei).
        :type value: int
        :param data: The transaction data, defaults to `0x`.
        :type data: HexStr

        Keyword Args:
            gas_price (int): Specify the gas price for the transaction.
            gas_limit (int): Specify the maximum gas the transaction can use.
            access_list (List[Dict]): Specify a list of addresses and storage
                keys that the transaction plans to access (EIP-2930).
            wait_for_receipt (bool): Wait for the transaction receipt,
                defaults to :const:`True`.

        :returns: The transaction receipt or the transaction hash if
            `wait_for_receipt` is :const:`False`.
        :rtype: Union[TxReceipt, HexStr]
        """

        if to:
            to = Web3.to_checksum_address(to)

        tx = {
            "from": self.address,
            "to": to,
            "chainId": self.w3.eth.chain_id,
            "nonce": self.w3.eth.get_transaction_count(self.address),
            "value": value,
            "gasPrice": kwargs.get("gas_price", self.w3.eth.gas_price),
            "data": data,
        }
        if kwargs.get("access_list"):
            tx["accessList"] = kwargs["access_list"]
        try:
            estimate_gas = self.w3.eth.estimate_gas(tx) + GAS_BUFFER
        except Exception:
            estimate_gas = 3000000
        tx["gas"] = kwargs.get("gas_limit", estimate_gas)
        tx = self.eth_acct.sign_transaction(tx).rawTransaction
        tx_hash = self.w3.eth.send_raw_transaction(tx).hex()
        logger.info(f"Transaction to {to}: {tx_hash}")
        if not kwargs.get("wait_for_receipt", True):
            return tx_hash
        receipt = self.w3.eth.wait_for_transaction_receipt(tx_hash)
        if not receipt.status:
            raise Exception(f"Transact to {to} failed.")
        return receipt



import os
import json
from typing import List, Tuple, Dict, Union, Any, Iterable
from hexbytes import HexBytes
from itertools import accumulate

from web3 import Web3
from web3.exceptions import MismatchedABI
from eth_typing import HexStr
import eth_abi
import rlp

from solcx import compile_source, set_solc_version
from solcx.install import install_solc
from solcx.exceptions import SolcNotInstalled

from cheb3.constants import TYPE_ALIAS


def load_compiled(contract_file: str, contract_name: str = None, base_path: str = "out/") -> Tuple[Dict, str]:
    """Loads compiled contracts from the project.

    :param contract_file: The name of the contract file.
    :type contract_file: str
    :param contract_name: Specifies the contract to be loaded. Defaults
        to the contract filename without suffix.
    :type contract_name: str
    :param base_path: Use the given path as the root of the source tree
        to load the compiled output. Defaults to "out/".
    :type base_path: str

    :return: ABI and bytecode of the required contract.
    :rtype: Tuple[Dict, str]
    """

    contract_name = contract_name or os.path.splitext(contract_file)[0]
    with open(os.path.join(base_path, contract_file, f"{contract_name}.json"), "r") as f:
        compiled = json.load(f)
        return (compiled["abi"], compiled["bytecode"]["object"])


def compile_file(
    contract_file: str,
    contract_names: Union[str, List[str]] = None,
    solc_version: str = None,
    base_path: str = None,
) -> Dict[str, Tuple[Dict, str]]:
    """Compile the Solidity source in the given file.

    Check :func:`compile_sol` for more details.
    """
    return compile_sol(
        open(contract_file, "r", encoding="utf-8").read(),
        contract_names=contract_names,
        solc_version=solc_version,
        base_path=base_path,
    )


def compile_sol(
    contract_source: str,
    contract_names: Union[str, List[str]] = None,
    solc_version: str = "latest",
    base_path: str = None,
) -> Dict[str, Tuple[Dict, str]]:
    """Compile the Solidity source and return the ABI and bytecode of
    the specific contracts.

    :param contract_source: The Solidity source code.
    :type contract_source: str
    :param contract_name: A target contract name or a list of target
        contract names, defaults to :const:`None`. If not given, it
        will return all contracts in the source file.
    :type contract_name: str | List[str]
    :param solc_version: `solc` version to use, defaults to :const:`latest`,
        i.e. the latest available version. If the specified version
        is not installed, it will be installed automatically.
    :type solc_version: str
    :param base_path: Use the given path as the root of the source tree
        to include other dependence contracts, e.g. the path to
        openzeppelin contracts. Defaults to :const:`None`.
    :type base_path: str

    :return: A dict, mapping the contract name to a tuple of the ABI and
        bytecode.
    :rtype: Dict[str, Tuple[Dict, str]]
    """
    if solc_version == "latest":
        solc_version = install_solc()

    try:
        set_solc_version(solc_version)
    except SolcNotInstalled:
        install_solc(solc_version)
        set_solc_version(solc_version)

    compiled = compile_source(
        contract_source,
        output_values=["abi", "bin"],
        solc_version=solc_version,
        base_path=base_path,
    )
    contracts = dict()
    if contract_names is None:
        contract_names = [c.split(":")[1] for c in compiled.keys() if c.startswith("<stdin>:")]
    if isinstance(contract_names, str):
        contract_names = [contract_names]
    for cn in contract_names:
        if f"<stdin>:{cn}" not in compiled:
            raise Exception(f"Contract {cn} not found.")
        contracts[cn] = (
            compiled[f"<stdin>:{cn}"]["abi"],
            compiled[f"<stdin>:{cn}"]["bin"],
        )
    return contracts


def decode_data(encoded_data: Union[HexBytes, HexStr], types: Iterable[str]) -> Union[Any, Tuple[Any]]:
    r"""The same as `abi.decode` in Solidity.

    Examples:

        >>> # decode a single value
        >>> decode_data(b'\xd0!\xb0\xc3\x07\xaf\x00\x9b?\xbe\x03\x99M\xb4\xfa\x9fy\x17 \x84'.rjust(32, b'\0'), ['address'])
        '0xd021b0c307af009b3fbe03994db4fa9f79172084'
        >>> # decode multiple values
        >>> decode_data('000000000000000000000000000000000000000000000000058d15e17628\
        00000000000000000000000000000000000000000000000000000000000000000064000000000\
        00000000000000000000000000000000000000000000000645f7142',\
        ('uint112', 'uint112', 'uint32'))
        (400000000000000000, 100, 1683976514)


    :param encoded_data: The encoded data.
    :type encoded_data: Union[~hexbytes.main.HexBytes, HexStr]
    :param types: A list or tuple of string representations of
        the ABI types that will be used for decoding.
    :type types: Iterable[str]

    :return: The decoded data.
    :rtype: Union[Any, Tuple[Any]]
    """

    def dfs(decoded_data: Tuple[Any], types: Union[str, Iterable[str]]) -> Tuple[Any]:
        if isinstance(types, str):
            types = [types]

        decoded_data = list(decoded_data)
        for i in range(len(types)):
            if "(" != types[i][0]:  # not struct
                if types[i].startswith("address"):
                    if isinstance(decoded_data[i], str):
                        decoded_data[i] = Web3.to_checksum_address(decoded_data[i])
                    else:
                        decoded_data[i] = tuple([Web3.to_checksum_address(addr) for addr in decoded_data[i]])
                continue

            if (square_left := types[i].rfind("[")) != -1:  # struct array
                decoded_data[i] = list(decoded_data[i])
                base = types[i][:square_left][1:-1]
                levels = accumulate((p == "(") - (s == ")") for p, s in zip(f" {base}", f"{base} "))
                _types = "".join([c, "\n"][c == "," and lv == 0] for c, lv in zip(base, levels)).split("\n")
                for j in range(len(decoded_data[i])):
                    decoded_data[i][j] = dfs(decoded_data[i][j], _types)
                decoded_data[i] = tuple(decoded_data[i])
            else:
                type_str = types[i][1:-1]
                levels = accumulate((p == "(") - (s == ")") for p, s in zip(f" {type_str}", f"{type_str} "))
                _types = "".join([c, "\n"][c == "," and lv == 0] for c, lv in zip(type_str, levels)).split("\n")
                decoded_data[i] = dfs(decoded_data[i], _types)

        return tuple(decoded_data)

    if isinstance(encoded_data, str):
        encoded_data = HexBytes(encoded_data)

    decoded = dfs(eth_abi.decode(types, encoded_data), types)
    if len(decoded) == 1:
        return decoded[0]
    return decoded


def encode_with_signature(signature: str, *args) -> HexStr:
    """The same as `abi.encodeWithSignature` in Solidity except that
    it can handle type alias.

    Examples:

        >>> encode_with_signature("transfer(address,uint)", "0x617F2E2fD72FD9D5503197092aC168c91465E7f2", 100)
        '0xa9059cbb000000000000000000000000617f2e2fd72fd9d5503197092ac168c91465e7f20000000000000000000000000000000000000000000000000000000000000064'

    :param signature: The function signature.
    :type signature: str
    :param `*args`: The parameters to be encoded.

    :return: The encoded data.
    :rtype: HexStr
    """

    def dfs(type_str: str) -> Tuple[str, Any]:
        if not type_str:
            return ("", [])
        levels = accumulate((p == "(") - (s == ")") for p, s in zip(f" {type_str}", f"{type_str} "))
        types = "".join([c, "\n"][c == "," and lv == 0] for c, lv in zip(type_str, levels)).split(
            "\n"
        )  # split by comma at level 0
        sig = ""
        for i in range(len(types)):
            if (square_left := types[i].rfind("[")) != -1:  # array
                base, append = types[i][:square_left], types[i][square_left:]
            else:
                base, append = types[i], ""
            if "(" == types[i][0]:
                ret = dfs(base[1:-1])
                sig += f"({ret[0]}){append},"
                types[i] = f"({','.join(ret[1])}){append}"
                continue
            types[i] = TYPE_ALIAS.get(base, base) + append
            sig += f"{types[i]},"
        return (sig[:-1], types)

    ret = dfs(signature[signature.find("(") + 1: -1])
    signature = signature[: signature.find("(") + 1] + ret[0] + ")"
    types = ret[1]
    if len(types) != len(args):
        raise MismatchedABI("Supplied parameters do not match the signature.")

    selector = Web3.solidity_keccak(["string"], [signature])[:4]

    parameters = eth_abi.encode(types, args)
    return f"0x{(selector + parameters).hex()}"


def calc_create_address(sender: HexStr, nonce: int) -> HexStr:
    """Calculate the address of the contract created by the given sender
    using the `CREATE` opcode with the given nonce.

    :param sender: The address of the sender.
    :type sender: HexStr
    :param nonce: The transaction count of the sender before the
        creation.
    :type nonce: int

    :return: The address of the contract.
    :rtype: HexStr
    """
    return Web3.to_checksum_address(Web3.keccak(rlp.encode([Web3.to_bytes(hexstr=sender), nonce]))[12:].hex())


def calc_create2_address(sender: HexStr, salt: int, initcode: HexStr) -> HexStr:
    """Calculate the address of the contract created by the given sender
    using the `CREATE2` opcode with the given salt and contract bytecode.

    :param sender: The address of the sender.
    :type sender: HexStr
    :param salt: The salt.
    :type salt: int
    :param initcode: The contract bytecode.
    :type initcode: HexStr

    :return: The address of the contract.
    :rtype: HexStr
    """
    return Web3.to_checksum_address(
        Web3.solidity_keccak(
            ["bytes1", "address", "uint256", "bytes32"],
            [
                b"\xff",
                Web3.to_checksum_address(sender),
                salt,
                Web3.solidity_keccak(["bytes"], [initcode]),
            ],
        )[12:].hex()
    )



from typing import Any
from hexbytes import HexBytes
from requests.exceptions import ConnectionError

from web3 import Web3
from web3.middleware import geth_poa_middleware

from cheb3.account import Account
from cheb3.contract import Contract


class Connection:
    """Creates a connection to an HTTP provider.

    :param endpoint_uri: The full URI to the RPC endpoint.
    :type endpoint_uri: str
    """

    def __init__(self, endpoint_uri: str) -> None:
        self.w3 = Web3(Web3.HTTPProvider(endpoint_uri))
        self.w3.middleware_onion.inject(geth_poa_middleware, layer=0)

        try:
            self.w3.is_connected(show_traceback=True)
        except Exception as e:
            if isinstance(e, ConnectionError):
                raise Exception(f"Could not connect to {endpoint_uri}.")
            # HTTPError('400 Client Error: Bad Request for url'): connected but is_connected() returns False

    def account(self, private_key: str = None) -> Account:
        """Creates an account associated with this connection.

        :param private_key: The private key of the account, defaults
            to :const:`None`. A new account will be created if not
            provided.
        :type private_key: str

        :rtype: :class:`Account <cheb3.account.Account>`
        """
        account_factory = Account.factory(self.w3)
        return account_factory(private_key)

    def contract(
        self,
        signer: Account = None,
        contract_name: str = "",
        address: str = None,
        **kwargs: Any,
    ) -> Contract:
        """Creates a contract instance associated with this connection.

        :param signer: The account that will sign the following transactions
            interacting with the contract, defaults to :const:`None`.
        :type signer: :class:`<cheb3.account.Account>`
        :param contract_name: The name of the contract, defaults to
            :const:`""`.
        :type contract_name: str
        :param address: Set the address of the contract if it is deployed,
            defaults to :const:`None`.

        Keyword Args:
            abi (ABI): The ABI of the contract.
            bytecode (HexStr): The bytecode of the contract.

        :rtype: :class:`Contract <cheb3.contract.Contract>`
        """
        contract_factory = Contract.factory(self.w3, contract_name)
        return contract_factory(signer, address, **kwargs)

    def get_balance(self, address: str) -> int:
        """Returns the balance of the given account.

        :param address: The address of the account.
        :type address: str

        :rtype: int
        """
        return self.w3.eth.get_balance(address)

    def get_storage_at(self, address: str, slot: int) -> HexBytes:
        """Returns the value from a storage position for the given account.

        :param address: The address of the account.
        :type address: str
        :param slot: The storage slot.
        :type slot: int

        :rtype: ~hexbytes.main.HexBytes
        """
        return self.w3.eth.get_storage_at(address, slot)

    def get_code(self, address: str) -> HexBytes:
        """Returns the code at the given account.

        :param address: The address of the account.
        :type address: str

        :rtype: ~hexbytes.main.HexBytes
        """
        return self.w3.eth.get_code(address)
